package gs.test.v0.execution.engine

import cats.data.Kleisli
import cats.effect.Async
import cats.syntax.all.*
import gs.test.v0.definition.TestGroupDefinition
import gs.test.v0.definition.TestSuite
import gs.test.v0.execution.SuiteExecution
import gs.test.v0.execution.TestExecution
import gs.timing.v0.Timing
import gs.uuid.v0.UUID
import java.time.Clock
import java.time.Instant
import natchez.EntryPoint
import natchez.Span

final class TestEngine[F[_]: Async](
  val configuration: EngineConfiguration,
  timing: Timing[F],
  suiteExecutionIdGenerator: UUID.Generator,
  testExecutionIdGenerator: UUID.Generator,
  clock: Clock,
  val entryPoint: EntryPoint[F]
):

  def runSuite(
    suite: TestSuite,
    tests: fs2.Stream[F, TestGroupDefinition[F]]
  ): F[SuiteExecution] =
    for
      executedAt <- Async[F].delay(Instant.now(clock))
      stats      <- EngineStats.initialize[F]
      _ <- tests
        .mapAsync(configuration.groupConcurrency.toInt())(runGroup)
        .evalTap(updateGroupStats)
        .evalTap(reportGroup)
        .flatMap(groupResult => fs2.Stream.emits(groupResult.testExecutions))
        .evalTap(updateTestStats)
        .evalMap(reportTestExecution)
        .compile
        .drain
      overallDuration <- stats.overallDuration.get
      countSeen       <- stats.countSeen.get
      countSucceeded  <- stats.countSucceeded.get
      countFailed     <- stats.countFailed.get
      countIgnored    <- stats.countIgnored.get
    yield SuiteExecution(
      id = suiteExecutionIdGenerator.next(),
      name = suite.name,
      documentation = suite.documentation,
      duration = overallDuration,
      countSeen = countSeen,
      countSucceeded = countSucceeded,
      countFailed = countFailed,
      countIgnored = countIgnored,
      executedAt = executedAt
    )

  private def updateGroupStats(groupResult: GroupResult): F[Unit] = ???

  private def updateTestStats(testExecution: TestExecution): F[Unit] = ???

  private def reportGroup(groupResult: GroupResult): F[Unit] = ???

  private def reportTestExecution(testExecution: TestExecution): F[Unit] = ???

  def runGroup(
    group: TestGroupDefinition[F]
  ): F[GroupResult] =
    for
      _      <- group.beforeGroup.getOrElse(Async[F].unit)
      stream <- executeGroupTests(group)
      _      <- group.afterGroup.getOrElse(Async[F].unit)
    yield stream

  private def executeGroupTests(group: TestGroupDefinition[F]): F[GroupResult] =
    for
      timer      <- timing.start()
      _          <- group.beforeGroup.getOrElse(Async[F].unit)
      executions <- streamGroupTests(group).compile.toList
      _          <- group.afterGroup.getOrElse(Async[F].unit)
      elapsed    <- timer.checkpoint()
    yield new GroupResult(
      name = group.name,
      documentation = group.documentation,
      duration = elapsed.duration,
      testExecutions = executions
    )

  private def streamGroupTests(group: TestGroupDefinition[F])
    : fs2.Stream[F, TestExecution] =
    fs2.Stream
      .emits(group.tests)
      .mapAsync(configuration.testConcurrency.toInt()) { test =>
        for
          testExecutionId <- Async[F].delay(testExecutionIdGenerator.next())
          timer           <- timing.start()
          _               <- group.beforeEachTest.getOrElse(Async[F].unit)
          result          <- test.unitOfWork
          _               <- group.afterEachTest.getOrElse(Async[F].unit)
          elapsed         <- timer.checkpoint()
        yield TestExecution(
          id = TestExecution.Id(testExecutionId),
          permanentId = test.permanentId,
          documentation = test.documentation,
          tags = test.tags,
          markers = test.markers,
          result = result,
          traceId = ???,
          sourcePosition = test.sourcePosition,
          duration = elapsed.duration
        )
      }
